/*
 * url2gr.c
 *
 * Convert <A HREF=url>title</A> lines to graph commands.
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

/*----------------------------------------------------------------------------*/

/*
 * current set of valid URL prefixes
 */
static char* prefixes[] = {
   "http://", "ftp://", "mailto:", "file:", "gopher:", 0 };

/*----------------------------------------------------------------------------*/

static char* OfficialName( char* host )
{
   static char official[128];
   struct hostent* hp = gethostbyname( host );
   if( hp )
   {
      strcpy( official, hp->h_name );
      free( hp );
   }
   else
      strcpy( official, host );
   return official;
}

/*----------------------------------------------------------------------------*/

static int ChopURL( char* s, char* t, char* h, char* p, char* f )
{
   if( sscanf( s, "%[^:]://%[^:]:%[^/]/%s", t, h, p, f ) == 4 ) return 1;
   if( sscanf( s, "%[^:]://%[^:]:%[^/]", t, h, p ) == 3 )
   {
      f[0] = 0;
      return 1;
   }
   if( sscanf( s, "%[^:]://%[^/]/%s", t, h, f ) == 3 )
   {
      p[0] = 0;
      return 1;
   }
   if( sscanf( s, "%[^:]://%[^/]", t, h  ) == 2 )
   {
      p[0] = 0;
      f[0] = 0;
      return 1;
   }
   return 0;
}

void GlueURL( char s[], char* t, char* h, char* p, char* f )
{
   if( p[0] && f[0] )
      sprintf( s, "%s://%s:%s/%s", t, h, p, f );
   else if( ! p[0] && f[0] )
      sprintf( s, "%s://%s/%s", t, h, f );
   else if( p[0] && ! f[0] )
      sprintf( s, "%s://%s:%s/", t, h, p );
   else if( ! p[0] && ! f[0] )
      sprintf( s, "%s://%s/", t, h );
}

/*----------------------------------------------------------------------------*/

/*
 * remove .. in URLs
 */
static void StripDots( char* s )
{
   int i;
   for( i = strlen(s)-1; i ; i-- )
      if( s[i] == '.' && s[i+1] == '.' )
      {
         int j;
         for( j = i-2 ; j >= 0; j-- )
            if( s[j] == '/' )
               break;
         strcpy( s+j+1, s+i+3 );
      }
}

/*
 * remove #... in URLs
 */
static void StripOffsets( char* s )
{
   int i;
   for( i = 0; i < strlen(s) ; i++ )
      if( s[i] == '#' )
      {
         s[i] = 0;
         break;
      }
}

/*----------------------------------------------------------------------------*/

/*
 * Check if URL has standard prefix.
 */
static int StdType( char* s )
{
   int i;
   for( i = 0; prefixes[i] != 0; i++ )
      if( strncmp( prefixes[i], s, strlen( prefixes[i] ) ) == 0 ) return i;
   return -1;
}

/*----------------------------------------------------------------------------*/

int main( int argc, char* argv[] )
{
   char s[512], base[512];
   char type[256], host[256], process[256], file[256];
   int first = 1;

   ChopURL( argv[1], type, host, process, file );
   file[0] = 0;
   GlueURL( base, type, host, process, file );

   while( gets( s ) != 0 )
   {
      char url[512], bFirstUrl[512], title[512];
      int n = sscanf( s, "<A HREF=\"%[^\"]\">%[^<]</A>", url, title );
      if( n == 2 )
      {
         char ur[512], *firstUrl;
         if( url[0] == '-' && url[1] == 0 )
            sprintf( ur, "%s", argv[1] );		/* base */
         else if( url[0] == '-' && url[1] != 0 )
            sprintf( ur, "%s%s", argv[1], url+1 );	/* ? */
         else if( url[0] == '/' && url[1] == 0 )
            sprintf( ur, "%s", base );			/* abs local */
         else if( url[0] == '/' && url[1] != 0 )
            sprintf( ur, "%s%s", base, url+1 );		/* abs local */
         else
         {
            int rc = ChopURL( url, type, host, process, file );
            if( rc )
               sprintf( ur, "%s", url );		/* global */
            else
               sprintf( ur, "%s%s", argv[1], url );	/* rel local */
         }
         StripDots( ur );
         StripOffsets( ur );

         {
            char* official;
            ChopURL( ur, type, host, process, file );
            official = OfficialName( host );
            GlueURL( ur, type, official, process, file );
         }

         printf( "v {%s} {%s}\n", ur, title );
         printf( "p v {%s} 4 bitmap %d\n", ur, StdType( ur ) + 1 );
         printf( "p v {%s} attr 4 -1\n", ur );

         if( first )
         {
            int i;
            firstUrl = (char*)strdup( ur );
            sprintf( bFirstUrl, "%s", ur );
            for( i = strlen(ur)-1; i ; i-- ) if( ur[i] == '/' ) break;
            bFirstUrl[i] = 0;
            first = 0;
         }
         else
            printf( "a {%s,%s} {%s} {%s} link\n", firstUrl, ur, firstUrl, ur );
      }
   }
   return 0;
}

/*----------------------------------------------------------------------------*/
