#include <stdio.h>
#include <assert.h>
#include <tk.h>

#include <hygraph/Hygraph.h>
#include <util/MN_Series.h>
#include <util/MN_Dictionary.h>
#include <util/MN_Set.h>
#include <util/MN_Color.h>
#include <util/unix.h>
#include <server/server.h>
#include <server/hgInfo.h>

#include "misc.h"
#include "util.h"
#include "tclTkHInfo.h"

#ifdef VER_TK
#  include "tkHInfo.h"
//#  include "widget.h"
   TkHInfo* current_hi = NIL;
#else
#  include "tclHInfo.h"
   TclHInfo* current_hi = NIL;
#endif

//------------------------------------------------------------------------------

#ifdef VER_TK

void Destroy( Vertex* v, Tcl_Interp* interp )
{
   delete (DQInfo*)v->vp;

   sprintf( tmpbuf, ".canvas delete %d", v->c->widget );
   Tcl_GlobalEval( interp, tmpbuf );
   sprintf( tmpbuf, ".canvas delete %d",
      v->c->FindSubContainer( REG_L )->widget );
   Tcl_GlobalEval( interp, tmpbuf );
   sprintf( tmpbuf, ".canvas delete %d",
      v->c->FindSubContainer( REG_I )->widget );
   Tcl_GlobalEval( interp, tmpbuf );
}

void Destroy( AArc* a, Tcl_Interp* interp )
{
   delete (DQInfo*)a->vp;

   sprintf( tmpbuf, ".canvas delete %d", a->ca->widget );
   Tcl_GlobalEval( interp, tmpbuf );
}

void Destroy( Blob* b, Tcl_Interp* interp )
{
   delete (DQInfo*)b->vp;

   sprintf( tmpbuf, ".canvas delete %d",
      b->c->FindSubContainer( REG_L )->widget );
   Tcl_GlobalEval( interp, tmpbuf );
   sprintf( tmpbuf, ".canvas delete %d",
      b->c->FindSubContainer( REG_NR )->widget );
   Tcl_GlobalEval( interp, tmpbuf );
}

void RaiseLower( HGContainer* C, Tcl_Interp* interp )
{
   HGContainer* c;
   if( C->widget != 0 )
   {
      sprintf( tmpbuf, ".canvas lower %d", C->widget );
      Tcl_GlobalEval( interp, tmpbuf );

      for( c = C->Parent(); c ; c = c->Parent() )
         if( c->widget != 0 ) break;

      if( c && c->widget != 0 )
         sprintf( tmpbuf, ".canvas raise %d %d", C->widget, c->widget );
      else
         sprintf( tmpbuf, ".canvas raise %d grid", C->widget );
      Tcl_GlobalEval( interp, tmpbuf );
   }
   for( c = C->Child(); c ; c = c->Sibling() )
      RaiseLower( c, interp );
}

#endif

//------------------------------------------------------------------------------

#ifdef VER_TK
COMMAND( grParm )
{
   Tcl_HashEntry *e;

// EGN: remove last 3 options:  all we need is grHandle [v|a|b] w Widget
//      and grHandle [v|a|b] o id ???
//
   CHK_CUR;
   CHK_ARGS( !=, 4, "usage: grParm [v|a|b] Widget [id|label|lwidget]" );

   int ty;
   if( (ty = Type( argv[1][0] )) < 0 ) RET_ERROR("bad type")

   e = Tcl_FindHashEntry( current_hi->htable[ty], (char*)atoi(argv[2]) );
   if( e )
   {
      HGObject* o = (HGObject*)Tcl_GetHashValue( e );
      if( strcmp( argv[3], "id" ) == 0 )
         sprintf( interp->result, "%s", current_hi->hygraph->Id(o) );
      else if( strcmp( argv[3], "label" ) == 0 )
         sprintf( interp->result, "%s", current_hi->hygraph->Label(o) );
      else if( strcmp( argv[3], "lwidget" ) == 0 )
         if( ty != 1 )
            sprintf( interp->result, "%d",
               o->c->FindSubContainer( REG_L )->widget );
         else
            RET_ERROR("no arc label widget")
      return TCL_OK;
   }
   else
      RET_ERROR("no element")
}
#endif

//------------------------------------------------------------------------------

COMMAND( grCreate )
{
   CHK_ARGS( <, 3, "usage: grCreate [h|v|a|b|l|f|m] Id Parms" );
   switch( argv[1][0] )
   {
      case 'h':
         CHK_ARGS( !=, 3, "usage: grCreate h Id" );
#        ifdef VER_TK
            current_hi = TkAddH( argv[2], interp );
#        else
            current_hi = TclAddH( argv[2], interp );
#        endif
         break;

      case 'v':
         CHK_CUR;
#        ifdef VER_TK
            CHK_ARGS( <, 4, "usage: grCreate v Id Label [X Y]" );
#        else
            CHK_ARGS( !=, 4, "usage: grCreate v Id Label" );
#        endif

         if( current_hi->AddV( argv[2], argv[3] ) == ERR_OK )
         {
            // EGN: maybe create widgets here rather than in the Tcl code?
            //
            Vertex* v = current_hi->hygraph->Vid( argv[2] );
#           ifdef VER_TK
               HGContainer* ci = v->c->FindSubContainer( REG_I );
               HGContainer* cl = v->c->FindSubContainer( REG_L );

               sprintf( tmpbuf, "CreateNodeRegion" );
               Tcl_GlobalEval( interp, tmpbuf );
               v->c->widget = atoi( interp->result );

               sprintf( tmpbuf, "CreateLabel \"%s\"", argv[3] );
               Tcl_GlobalEval( interp, tmpbuf );
               sscanf( interp->result, "%d %g %g", &cl->widget,
                  &cl->size.x, &cl->size.y );
                cl->pos.Set(0,0,0);

               sprintf( tmpbuf, "CreateIconRegion" );
               Tcl_GlobalEval( interp, tmpbuf );
               sscanf( interp->result, "%d %g %g", &ci->widget,
                  &ci->size.x, &ci->size.y );
                ci->pos.Set(0,0,0);

               int rc;
               Tcl_HashEntry *e = Tcl_CreateHashEntry( current_hi->htable[0],
                  (char*)ci->widget, &rc );
               if( rc )
                  Tcl_SetHashValue( e, v );
               else
                  fprintf( stderr, "couldn't create hash entry!\n" );

               NopLayout l;
               l.DoIt( v->c );

               MN_Point p;
               if( argc >= 6 )
               {
                  p.x = Max( (float)0, atof( argv[4] ) - v->c->size.x / 2 );
                  p.y = Max( (float)0, atof( argv[5] ) - v->c->size.y / 2 );
               }
               else
               {
                  Tcl_GlobalEval( interp, "winfo width .canvas" );
                  float width = atof( interp->result ) - 10.0;
                  Tcl_GlobalEval( interp, "winfo height .canvas" );
                  float height = atof( interp->result ) - 10.0;

                  p.Set( Random(10.0,width), Random(10.0,height), 0 );
               }
               v->c->offset.Set(0,0,0);
               v->c->pos = p;
               v->c->ComputeOffsets();
               current_hi->DumpGeom( v->c, "", current_hi->geom[0], 0 );

#           endif
            v->vp = new DQInfo( 0 );
            return TCL_OK;
         }
         else
            RET_ERROR("bad vertex")
         break;

      case 'a':
         CHK_CUR;
         CHK_ARGS( !=, 6, "usage: grCreate a Id FromId ToId Label" );

         if( current_hi->AddA( argv[2], argv[3], argv[4], argv[5] ) == ERR_OK )
         {
            AArc* a = current_hi->hygraph->Aid( argv[2] );

#           ifdef VER_TK

               sprintf( tmpbuf, "CreateLink" );
               Tcl_GlobalEval( interp, tmpbuf );
               sscanf( interp->result, "%d", &a->ca->widget );

               current_hi->DumpMP( a->ca, "", a->ca->MP() );

               int rc;
               Tcl_HashEntry *e = Tcl_CreateHashEntry( current_hi->htable[1],
                  (char*)a->ca->widget, &rc );
               if( rc )
                  Tcl_SetHashValue( e, a );
               else
                  fprintf( stderr, "couldn't create hash entry!\n" );

#           endif
            a->vp = new DQInfo( 1 );
            return TCL_OK;
         }
         else
            RET_ERROR("bad arc")
         break;

      case 'b':
         CHK_CUR;
         CHK_ARGS( !=, 5, "usage: grCreate b Id VId Label" );

         if( current_hi->AddB( argv[2], argv[3], argv[4] ) == ERR_OK )
         {
            Blob* b = current_hi->hygraph->Bid( argv[2] );
#           ifdef VER_TK
               Vertex* v = current_hi->hygraph->Vid( argv[3] );
               HGContainer* cl = b->c->FindSubContainer( REG_L );
               HGContainer* cn = b->c->FindSubContainer( REG_NR ); 

               sprintf( tmpbuf, "CreateNodesRegion" );
               Tcl_GlobalEval( interp, tmpbuf );
               sscanf( interp->result, "%d %g %g", &cn->widget,
                  &cn->size.x, &cn->size.y );

               if( strlen( argv[4] ) > 0 )
               {
                  sprintf( tmpbuf, "CreateLabel \"%s\"", argv[4] );
                  Tcl_GlobalEval( interp, tmpbuf );
                  sscanf( interp->result, "%d %g %g", &cl->widget,
                     &cl->size.x, &cl->size.y );
               }
               else
                  cl->size.Set( 1, 1, 0 );

               int rc;
               Tcl_HashEntry *e = Tcl_CreateHashEntry( current_hi->htable[2],
                  (char*)cn->widget, &rc );
               if( rc )
                  Tcl_SetHashValue( e, b );
               else
                  fprintf( stderr, "couldn't create hash entry!\n" );

               RaiseLower( b->c, interp );

               GridLayout l;
               l.Draw( v->c );
               current_hi->DumpGeom( v->c, current_hi->hygraph->Id( v ),
                  current_hi->geom[0], 0 );

#           endif
            b->vp = new DQInfo( 2 );
            return TCL_OK;
         }
         else
            RET_ERROR("bad blob")
         break;

      case 'l':
         CHK_ARGS( !=, 4, "usage: grCreate l Id Layout" );
         return gsAddL( argv[2], argv[3] ) == ERR_OK ? TCL_OK : TCL_ERROR;
         break;

      case 'f':
         CHK_CUR;
         CHK_ARGS( !=, 4, "usage: grCreate f Id Mag" );
         {
            double m;
            if( Tcl_GetDouble(interp,argv[3],&m) != TCL_OK ) return TCL_ERROR;
            Vertex* v = current_hi->hygraph->Vid( argv[2] );
            if( v == NIL )
               RET_ERROR("bad vertex")
            current_hi->fpts.Add( v );
            DQ(mag,v) = m;
            return TCL_OK;
         }
         break;

      case 'm':
         CHK_CUR;
         CHK_ARGS( !=, 4, "usage: grCreate m Index Size" );
         {
            int i, s;
            if( Tcl_GetInt( interp, argv[2], &i ) != TCL_OK ) return TCL_ERROR;
            if( Tcl_GetInt( interp, argv[3], &s ) != TCL_OK ) return TCL_ERROR;
            return current_hi->CreateMatrix( i, s ) == ERR_OK
                   ? TCL_OK : TCL_ERROR;
         }
         break;

      case 'p':
         CHK_CUR;
         CHK_ARGS( !=, 9,
            "usage: grCreate p Index (forward|reverse|undirected) Rel VVar AVar VExpr AExpr" );
         {
            int i, d, rel, rc;
            if( Tcl_GetInt( interp, argv[2], &i ) != TCL_OK ) return TCL_ERROR;
            if(      strcmp( argv[3], "forward"    ) == 0 ) d = 0;
            else if( strcmp( argv[3], "reverse"    ) == 0 ) d = 1;
            else if( strcmp( argv[3], "undirected" ) == 0 ) d = 2;
            else RET_ERROR("bad direction")
            if( argv[4][0] == 'A' ) rel = 1;
            else if( argv[4][0] == 'B' ) rel = 2;
            else RET_ERROR("bad relation")
            rc = current_hi->InitMinWtPaths( i, (Direction)d, rel,
               argv[5], argv[6], argv[7], argv[8] );
            return rc != -1 ? TCL_OK : TCL_ERROR;
         }
         break;

      case 's':
      case 'e':
         CHK_CUR;
         CHK_ARGS( !=, 3, "usage: grCreate [s|e] Id" );
         {
            Vertex* v = current_hi->hygraph->Vid( argv[2] );
            if( v == NIL )
               RET_ERROR("bad vertex")
            if( argv[1][0] == 's' )
               current_hi->UpdateEndPt( v, true, 1 );
            else
               current_hi->UpdateEndPt( v, false, 1 );
            return TCL_OK;
         }
         break;

      case 'i':
         CHK_ARGS( !=, 6, "usage: grCreate i Id Small Normal Large" );
         sprintf( tmpbuf, "b2 %s %s %s %s", argv[2], argv[3], argv[4],
                  argv[5] );
         Tcl_GlobalEval( interp, tmpbuf );
         break;

      default:
         RET_ERROR("invalid type")
         break;
   }
   return TCL_OK;
}

//------------------------------------------------------------------------------

COMMAND( grDestroy )
{
   CHK_CUR;
   CHK_ARGS( !=, 3, "usage: grDestroy [h|v|a|b|l|f|s|e] Id" );

   switch( argv[1][0] )
   {
      case 'v':
      {
         Vertex* v = current_hi->hygraph->Vid( argv[2] );
         if( v != NIL )
         {
#           ifdef VER_TK
               int i;
               AArc* a;
               Blob* b;
               Destroy( v, interp );
               for( i = 0; (b = current_hi->hygraph->B(v,i)) != 0; i++ )
                  Destroy( b, interp );
               for( i = 0; (a = current_hi->hygraph->In(v,i)) != 0; i++ )
                  Destroy( a, interp );
               for( i = 0; (a = current_hi->hygraph->Out(v,i)) != 0; i++ )
                  Destroy( a, interp );
#           endif

            return current_hi->Remove( v ) == ERR_OK ? TCL_OK : TCL_ERROR;
         }
         else
            RET_ERROR("bad vertex")
         break;
      }
      case 'a':
      {
         AArc* a = current_hi->hygraph->Aid( argv[2] );
         if( a != NIL )
         {
#           ifdef VER_TK
               Destroy( a, interp );
#           endif
            return current_hi->Remove( a ) == ERR_OK ? TCL_OK : TCL_ERROR;
         }
         else
            RET_ERROR("bad arc")
         break;
      }
      case 'b':
      {
         Blob* b = current_hi->hygraph->Bid( argv[2] );
         if( b != NIL )
         {
#           ifdef VER_TK
               Destroy( b, interp );
#           endif
            return current_hi->Remove( b ) == ERR_OK ? TCL_OK : TCL_ERROR;
         }
         else
            RET_ERROR("bad blob")
         break;
      }
      case 'h':
      {
         HygraphInfo* i = gsFindHI( argv[2] );
         if( i != NIL )
         {
#           ifdef VER_TK
               Blob* b;
               AArc* a;
               Vertex* v;
               int j;

               for( j = 0; (b = current_hi->hygraph->B(j)) != 0; j++ )
                  Destroy( b, interp );
               for( j = 0; (a = current_hi->hygraph->A(j)) != 0; j++ )
                  Destroy( a, interp );
               for( j = 0; (v = current_hi->hygraph->V(j)) != 0; j++ )
                  Destroy( v, interp );
            #endif

            if( current_hi == i ) current_hi = NIL;
            return gsDel( i ) == ERR_OK ? TCL_OK : TCL_ERROR;
         }
         else
            RET_ERROR("bad hygraph")
         break;
      }
      case 'l':
      {
         Layout* l = gsFindL( argv[2] );
         if( l != NIL )
         {
            return gsDel( l ) == ERR_OK ? TCL_OK : TCL_ERROR;
         }
         else
            RET_ERROR("bad layout")
         break;
      }
      case 'f':
      {
         Vertex* v = current_hi->hygraph->Vid( argv[2] );
         if( v == NIL )
            RET_ERROR("focal point")
         current_hi->fpts.Remove( v );
         return TCL_OK;
         break;
      }
      case 's':
      case 'e':
      {
         Vertex* v = current_hi->hygraph->Vid( argv[2] );
         if( v == NIL )
            RET_ERROR("bad vertex")
         if( argv[1][0] == 's' )
            current_hi->UpdateEndPt( v, true, -1 );
         else
            current_hi->UpdateEndPt( v, false, -1 );
         return TCL_OK;
         break;
      }
      case 'i':
         sprintf( tmpbuf, "b2 %s {} {} {} 0", argv[2] );
         Tcl_GlobalEval( interp, tmpbuf );
         return TCL_OK;
         break;
      default:
         RET_ERROR("invalid type")
         break;
   }
}

//------------------------------------------------------------------------------

COMMAND( grConfigure )
{
   CHK_ARGS( <, 4, "usage: grConfigure [h|v|a|b|l|f|m] Id Property" );
   if( strcmp( argv[3], "layout") == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 6, "usage: grConfigure [h|b] Id layout Region LId" );
      int ty;
      if( (ty = Type( argv[1][0] )) < 0 ) RET_ERROR("bad type")

      int r;
      if( Tcl_GetInt( interp, argv[4], &r ) != TCL_OK ) return TCL_ERROR;
      return current_hi->SetLayout( ty, argv[2], r, argv[5] ) == ERR_OK
             ? TCL_OK : TCL_ERROR;
   }
   else if( strcmp( argv[3], "attr" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( <, 6, "usage: grConfigure [v|a|b] Id attr Index Expr ..." );

      int t = Type( argv[1][0] );
      if( ! InRange( t, 0, 2 ) ) RET_ERROR("bad type");
      HGObject* o = t == 0 ? current_hi->hygraph->Vid( argv[2] ) :
                    t == 1 ? current_hi->hygraph->Aid( argv[2] ) :
                             current_hi->hygraph->Bid( argv[2] );
      if( o != NIL )
      {
         // Fake loop context by pushing a temporary variable on the stack.
         //
         char value[4], name[4];
         //sprintf( value, "%d", stack.last );
         sprintf( value, "%d", (int)o );
         sprintf( name, "%c", argv[1][0] );
         Tcl_SetVar( interp, name, value, 0 );
         //stack.Push( t );
         //stack.Set( o );

         for( int j = 4; j < argc; j += 2 )
         {
            int i;
            if( Tcl_GetInt( interp, argv[j], &i ) != TCL_OK ) return TCL_ERROR;
            if( ! InRange( i, 0, 31 ) ) RET_ERROR("invalid index")

            //DQ(val[i],o) = atof( argv[j+1] );

            double d;
            Tcl_ExprDouble( interp, argv[j+1], &d );
            DQ(val[i],o) = (float)d;

            current_hi->sdirty[t][i] = 1;
         }

         // Restore stack and delete temporary variable.
         //
         //stack.Pop();
         Tcl_UnsetVar( interp, name, 0 );

         return TCL_OK;
      }
      RET_ERROR("bad object")
   }
   else if( strcmp( argv[3], "label" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( <, 4, "usage: grConfigure [v|a|b] Id label [Label]" );

      if( argc == 4 )
      {
         switch( argv[1][0] )
         {
            case 'v':
            {
               Vertex* v = current_hi->hygraph->Vid( argv[2] );
               if( v == NIL ) RET_ERROR("bad vertex")
               sprintf( interp->result, "%s", current_hi->hygraph->Label(v) );
               break;
            }
            case 'a':
            {
               AArc* a = current_hi->hygraph->Aid( argv[2] );
               if( a == NIL ) RET_ERROR("bad arc")
               sprintf( interp->result, "%s", current_hi->hygraph->Label(a) );
               break;
            }
            case 'b':
            {
               Blob* b = current_hi->hygraph->Bid( argv[2] );
               if( b == NIL ) RET_ERROR("bad blob")
               sprintf( interp->result, "%s", current_hi->hygraph->Label(b) );
               break;
            }
            default:
               RET_ERROR("bad type")
               break;
         }
      }
      else
      {
         HGObject* o = NIL;
         int w = 0;
         switch( argv[1][0] )
         {
            case 'v':
               o = current_hi->hygraph->Vid( argv[2] );
               if( o == NIL ) RET_ERROR("bad vertex")
               w = ( o->c->FindSubContainer( REG_L ) )->widget;
               break;
            case 'a':
               o = current_hi->hygraph->Aid( argv[2] );
               if( o == NIL ) RET_ERROR("bad arc")
               w = ((AArc*)o)->cl->widget;
               break;
            case 'b':
               o = current_hi->hygraph->Bid( argv[2] );
               if( o == NIL ) RET_ERROR("bad blob")
               w = ( o->c->FindSubContainer( REG_L ) )->widget;
               break;
            default:
               RET_ERROR("bad type")
               break;
         }
         if( w != 0 )
         {
            current_hi->hygraph->Label( o, argv[4] );
            sprintf( tmpbuf, ".canvas itemconfig %d -text {%s}", w, argv[4] );
            Tcl_GlobalEval( interp, tmpbuf );
         }
         current_hi->hygraph->Label( o, argv[4] );
      }
      return TCL_OK;
   }
   else if( strcmp( argv[3], "bbox" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 6, "usage: grConfigure [h|v|b] Id bbox Region {X Y U V}" );

      int r;
      int ty;
      if( (ty = Type( argv[1][0] )) < 0 ) RET_ERROR("bad type")
      if( Tcl_GetInt( interp, argv[4], &r ) != TCL_OK ) return TCL_ERROR;
      HGContainer* c =
         current_hi->Container( ty, argv[2], r );

      if( c == NIL ) RET_ERROR("bad element or invalid region")
      else
      {
         MN_Point p0, p1;
         float x, y, u, v;
         if( sscanf( argv[5], "%f %f %f %f", &x, &y, &u, &v ) != 4 )
            RET_ERROR("bad bbox");

         p0.Set( x, y, 0 );
         p1.Set( u - x, v - y, 0 );
         c->SetPos( p0 );
         c->SetSize( p1 );

         //  Back up hierarchy to top-most node, redraw, and then output
         //  new geometry.
         //
/*
         for( HGContainer* cc = c; cc->Parent() ; cc = cc->Parent() )
            if( cc->Parent()->key == REG_H ) break;

         NopLayout l;
         l.DoIt( cc );
#        ifdef VER_TK
            current_hi->Redisplay( cc );
#        endif
*/

         return TCL_OK;
      }
   }
   else if( strcmp( argv[3], "pos" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 7, "usage: grConfigure [v|b] Id pos Region X Y" );

      int ty;
      if( (ty = Type( argv[1][0] )) < 0 ) RET_ERROR("bad type")

      int r;
      if( Tcl_GetInt( interp, argv[4], &r ) != TCL_OK ) return TCL_ERROR;
      HGContainer* c =
         current_hi->Container( ty, argv[2], r );

      if( c == NIL ) RET_ERROR("bad element or invalid region")
      else
      {
         MN_Point p;
         double d;
         if( Tcl_GetDouble(interp,argv[5],&d) != TCL_OK ) return TCL_ERROR;
         p.x = d;
         if( Tcl_GetDouble(interp,argv[6],&d) != TCL_OK ) return TCL_ERROR;
         p.y = d;
         c->SetPos( p );
         return TCL_OK;
      }
   }
#ifdef VER_TK
   else if( strcmp( argv[3], "movenode" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 6, "usage: grConfigure [v|b] Id movenode DX DY" );

      Vertex* v = current_hi->hygraph->Vid( argv[2] );
      if( v == NIL ) RET_ERROR("bad vertex")

      {
         MN_Point p;
         double d;
         if( Tcl_GetDouble(interp,argv[4],&d) != TCL_OK ) return TCL_ERROR;
         p.x = d;
         if( Tcl_GetDouble(interp,argv[5],&d) != TCL_OK ) return TCL_ERROR;
         p.y = d;
         current_hi->Move( v, &p );
         return TCL_OK;
      }
   }
#endif
   else if( strcmp( argv[3], "size" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 5, "usage: grConfigure m Index size Size" );
      if( argv[1][0] != 'm' ) RET_ERROR("invalid type")
      int i, s;
      if( Tcl_GetInt( interp, argv[2], &i ) != TCL_OK ) return TCL_ERROR;
      if( Tcl_GetInt( interp, argv[4], &s ) != TCL_OK ) return TCL_ERROR;
      return current_hi->EnlargeMatrix( i, s ) == ERR_OK ? TCL_OK : TCL_ERROR;
   }
/*
   else if( strcmp( argv[3], "path" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 4, "usage: grConfigure p Index path" );
      if( argv[1][0] != 'p' ) RET_ERROR("invalid type")
      int i;
      if( Tcl_GetInt( interp, argv[2], &i ) != TCL_OK ) return TCL_ERROR;
      return current_hi->MinWtPaths( i ) == ERR_OK ? TCL_OK : TCL_ERROR;
   }
*/
   else if( strcmp( argv[3], "regexp" ) == 0 )
   {
      char f = argv[4][0];
      int ty;
      if( (ty = Type( argv[1][0] )) < 0 ) RET_ERROR("bad type")
      int i;
      CHK_CUR;
      CHK_ARGS( !=, 6,
         "usage: grConfigure [v|a|b] Index regexp [i|l] RegExpr" );
      if( Tcl_GetInt( interp, argv[2], &i ) != TCL_OK ) return TCL_ERROR;
      if( f != 'i' && f != 'l' ) RET_ERROR("invalid field")
      return current_hi->RegExpr( ty, f == 'i', argv[5], i ) == ERR_OK
             ? TCL_OK : TCL_ERROR;
   }
   else if( strcmp( argv[3], "icon" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 5, "usage: grConfigure v Id icon IconId" );
      if( argv[1][0] != 'v' ) RET_ERROR("invalid type")
      Vertex* v = current_hi->hygraph->Vid( argv[2] );
      if( v == NIL ) RET_ERROR("bad vertex")
      sprintf( tmpbuf, "BtoN %d %s",
         v->c->FindSubContainer( REG_I )->widget, argv[4] );
      Tcl_GlobalEval( interp, tmpbuf );
      return TCL_OK;
   }
   else if( strcmp( argv[3], "arrow" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 5, "usage: grConfigure a Id arrow [none|first|last|both]" );
      if( argv[1][0] != 'a' ) RET_ERROR("invalid type")
      sprintf( tmpbuf,
         ".canvas itemconfigure $linkFromId(%s) -arrow %s", argv[2], argv[4] );
      Tcl_GlobalEval( interp, tmpbuf );
      return TCL_OK;
   }
   else if( strcmp( argv[3], "blob" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( <, 4, "usage: grConfigure v Id blob bId" );
      if( argv[1][0] != 'v' ) RET_ERROR("invalid type")
      Vertex* v = current_hi->hygraph->Vid( argv[2] );
      if( v == NIL ) RET_ERROR("bad vertex")
      Blob* b = NIL;
      if( argc >= 5 )
      {
         b = current_hi->hygraph->Bid( argv[4] );
         if( b == NIL ) RET_ERROR("bad blob")
      }
      current_hi->hygraph->MoveVertex2Blob( v, b );

#     ifdef VER_TK
         if( b != NIL )
         {
            HGContainer* cvl = v->c->FindSubContainer( REG_L );
            HGContainer* cvi = v->c->FindSubContainer( REG_I );
            HGContainer* cbn = b->c->FindSubContainer( REG_NR );
            sprintf(tmpbuf, ".canvas raise %d %d", cvi->widget, cbn->widget );
            Tcl_GlobalEval( interp, tmpbuf );
            sprintf(tmpbuf, ".canvas raise %d %d", cvl->widget, cbn->widget );
            Tcl_GlobalEval( interp, tmpbuf );

            GridLayout l;
            l.Draw( b->c );
            current_hi->DumpGeom( b->c, "", current_hi->geom[2], 2 );
         }
         RaiseLower( v->c, interp );
#     endif

      return TCL_OK;
   }
   else
      RET_ERROR("invalid property")
}

//------------------------------------------------------------------------------

// grConf var bbox Region {X Y U V}
//
COMMAND( grConf )
{
   if( strcmp( argv[2], "bbox" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 5, "usage: grConf var bbox Region {X Y U V}" );

      int r;
      HGObject* o = (HGObject*)strtol( argv[1], (char**)NULL, 16 );
      if( Tcl_GetInt( interp, argv[3], &r ) != TCL_OK ) return TCL_ERROR;

      HGContainer* c = gsSContainer( o->c, r );

      if( c == NIL ) RET_ERROR("bad element or invalid region")
      else
      {
         MN_Point p0, p1;
         float x, y, u, v;
         if( sscanf( argv[4], "%f %f %f %f", &x, &y, &u, &v ) != 4 )
            RET_ERROR("bad bbox");

         p0.Set( x, y, 0 );
         p1.Set( u - x, v - y, 0 );
         c->SetPos( p0 );
         c->SetSize( p1 );

         //  Back up hierarchy to top-most node, redraw, and then output
         //  new geometry.
         //
/*
         for( HGContainer* cc = c; cc->Parent() ; cc = cc->Parent() )
            if( cc->Parent()->key == REG_H ) break;

         NopLayout l;
         l.DoIt( cc );
#        ifdef VER_TK
            current_hi->Redisplay( cc );
#        endif
*/
         return TCL_OK;
      }
   }
   else
      RET_ERROR("invalid property")
}

//------------------------------------------------------------------------------

#ifdef VER_TK
COMMAND( grScale )
{
   CHK_CUR;
   CHK_ARGS( !=, 2, "usage: grScale S" );
   float s = atof( argv[1] );
   MN_Point p( s, s, 1 );
   current_hi->Scale( &p );
   return TCL_OK;
}
#endif

//------------------------------------------------------------------------------

COMMAND( grDraw )
{
   int rc;
   CHK_CUR;
   CHK_ARGS( <, 2, "usage: grDraw [v|a] Sep" );
   switch( argv[1][0] )
   {
      case 'v':
         CHK_ARGS( !=, 2, "usage: grDraw v" );
         rc = current_hi->Exec();
#        ifdef VER_TCL
            TclFlush( "s\n" );
#        endif
         return rc == ERR_OK ? TCL_OK : TCL_ERROR;
         break;
      case 'a':
      {
         CHK_ARGS( !=, 3, "usage: grDraw a Sep" );
         float s = Max( (float)atof( argv[2] ), (float)0 );
         int rc = current_hi->Reroute( s );
#        ifdef VER_TCL
            TclFlush( "s\n" );
#        endif
         return rc == ERR_OK ? TCL_OK : TCL_ERROR;
         break;
      }
      default:
         RET_ERROR("bad type")
         break;
   }
}

//------------------------------------------------------------------------------

COMMAND( grInfo )
{
   CHK_ARGS( <, 2, "usage: grInfo [version|handle [v|a|b] id]" );

   if( strcmp( argv[1], "version" ) == 0 )
   {
#     ifdef VER_TK
         sprintf( interp->result, "grwish version %s, %s.",
         VERSION, __DATE__ );
#     else
         sprintf( interp->result, "grtclsh version %s, %s.",
         VERSION, __DATE__ );
#  endif
   }
   else if( strcmp( argv[1], "handle" ) == 0 )
   {
      CHK_CUR;
      CHK_ARGS( !=, 4, "usage: grInfo handle [v|a|b] id" );
      HGObject* o = 0;
      switch( argv[2][0] )
      {
         case 'v': o = current_hi->hygraph->Vid( argv[3] ); break;
         case 'a': o = current_hi->hygraph->Aid( argv[3] ); break;
         case 'b': o = current_hi->hygraph->Bid( argv[3] ); break;
         default : RET_ERROR("bad type")
      }
      sprintf( interp->result, "0x%lx", (long unsigned int)o );
   }
   else RET_ERROR("bad info subcommand")

   return TCL_OK;
}

//------------------------------------------------------------------------------

int GrBase_Init( Tcl_Interp* interp )
{
   if( gsInit() != ERR_OK ) RET_ERROR("could not initialize package")

   current_hi = NIL;

   Tcl_CreateCommand( interp, "grCreate",    grCreate,    (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grDestroy",   grDestroy,   (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grConfigure", grConfigure, (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grConf",      grConf,      (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grDraw",      grDraw,      (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grInfo",      grInfo,      (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);

#ifdef VER_TK
   Tcl_CreateCommand( interp, "grParm",      grParm,      (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grScale",     grScale,     (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
#endif

   return TCL_OK;
}

//------------------------------------------------------------------------------
