#include <stdio.h>
#include <tcl.h>
#include <util/MN_Point.h>
#include <util/unix.h>
#include "util.h"

//------------------------------------------------------------------------------

COMMAND( gxAttach )
{
   int f[4], s[4], ii[2], ij[2];

   CHK_ARGS( !=, 3, "usage: gxAttach {FromCoords} {ToCoords}" );

   if( sscanf( argv[1], "%d %d %d %d", &f[0], &f[1], &f[2], &f[3] ) != 4 )
      return TCL_ERROR;
   if( sscanf( argv[2], "%d %d %d %d", &s[0], &s[1], &s[2], &s[3] ) != 4 )
      return TCL_ERROR;

   ComputeAttachmentPoints( f, s, ii, ij );

   sprintf( interp->result, "%d %d %d %d", ii[0], ii[1], ij[0], ij[1] );

   return TCL_OK;
}

//------------------------------------------------------------------------------

COMMAND( gxCenter )
{
   float f[4];
   CHK_ARGS( !=, 2, "usage: gxCenter {Coords}" );

   if( sscanf( argv[1], "%f %f %f %f", &f[0], &f[1], &f[2], &f[3] ) != 4 )
      return TCL_ERROR;

   sprintf( interp->result, "%g %g",
      f[0] + ( f[2] - f[0] ) / 2.0, f[1] + ( f[3] - f[1] ) / 2.0 );

   return TCL_OK;
}

//------------------------------------------------------------------------------

COMMAND( gxHSVtoRGB )
{
   MN_Point hsv, rgb;
   CHK_ARGS( !=, 4, "usage: gxHSVtoRGB H S V" );

   hsv.x = atof( argv[1] );
   hsv.y = atof( argv[2] );
   hsv.z = atof( argv[3] );

   HSVtoRGB( &hsv, &rgb );

   sprintf( interp->result, "%g %g %g", rgb.x, rgb.y, rgb.z );
   return TCL_OK;
}


//------------------------------------------------------------------------------

int GrGfxs_Init( Tcl_Interp *interp )
{
   Tcl_CreateCommand( interp, "gxAttach",   gxAttach,   (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "gxCenter",   gxCenter,   (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "gxHSVtoRGB", gxHSVtoRGB, (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);
   return TCL_OK;
}

//------------------------------------------------------------------------------
