//
//  FILE:       tclHInfo.C
//  AUTHORS:    Emanuel G. Noik
//

#include <assert.h>
#include <tcl.h>

#include <hygraph/Hygraph.h>
#include <util/MN_Dictionary.h>
#include <util/MN_Set.h>
#include <server/io.h>
#include <server/server.h>
#include <server/hgInfo.h>
#include "tclTkHInfo.h"
#include "tclHInfo.h"

//------------------------------------------------------------------------------

static Buffer* outBuf = NIL;

//------------------------------------------------------------------------------

#define PrPt(P)	P.x, P.y, P.z
#define PL(S)	outBuf->Put( S, strlen( S ) );

//------------------------------------------------------------------------------

void TclFlush( const char* s )
{
   if( strcmp( s, "" ) != 0 )
      PL( s );
   outBuf->Flush();
}

//------------------------------------------------------------------------------

TclHInfo* TclAddH( const char* id, Tcl_Interp* interp )
{
   MN_String* s = new MN_String( id );
   TclHInfo* i = (TclHInfo*)hiDict->LookUp( s );

   if( i == NIL )
   {
      i = new TclHInfo( id, interp );
      i->hygraph->Validate( false );
      hiDict->Add( s, i );
   }
   else
      delete s;

   return i;
}

//------------------------------------------------------------------------------

TclHInfo::TclHInfo( const char* id, Tcl_Interp* interp )
: TclTkHInfo( id, interp )
{
   if( outBuf == NIL ) outBuf = new Buffer( 2048 );
}

//------------------------------------------------------------------------------

int TclHInfo::DumpBB( HGContainer *C, const char *id, int t, int region )
{
   MN_Point pos, size;
   char type = Code( t );

   C->GetAbsPos( pos );
   C->GetSize( size );

   if( size.Magnitude() > 0.0 )
   {
      sprintf( tmpbuf, "p %c \"%s\" %d bb %g %g %g %g\n",
               type, id, region, pos.x, pos.y,
               pos.x + size.x, pos.y + size.y );
      PL( tmpbuf );
   }

   return DumpGC( C, id, t, region, gc_mask );
}

//------------------------------------------------------------------------------

int TclHInfo::DumpGC( HGContainer* C, const char *id, int t, int region,
int m )
{
   int mask = C->gc.dirty & m;
   char type = Type( t );

   if( mask & gc_fill )
   {
      Clr( C->gc, gc_fill );
      sprintf( tmpbuf, "p %c \"%s\" %d gc fill %d\n",
      type, id, region, C->gc.fill );
      PL( tmpbuf );
   }

   if( mask & gc_fg_color )
   {
      Clr( C->gc, gc_fg_color );
      sprintf( tmpbuf, "p %c \"%s\" %d gc fg %g %g %g\n",
      type, id, region, PrPt(C->gc.fg_color) );
      PL( tmpbuf );
   }

   if( mask & gc_bg_color )
   {
      Clr( C->gc, gc_bg_color );
      sprintf( tmpbuf, "p %c \"%s\" %d gc bg %g %g %g\n",
      type, id, region, PrPt(C->gc.bg_color) );
      PL( tmpbuf );
   }

   if( mask & gc_d )
   {
      Clr( C->gc, gc_d );
      sprintf( tmpbuf, "p %c \"%s\" %d gc d %g %g %g\n",
      type, id, region, PrPt(C->gc.d) );
      PL( tmpbuf );
   }

   if( mask & gc_s )
   {
      Clr( C->gc, gc_s );
      sprintf( tmpbuf, "p %c \"%s\" %d gc s %g %g %g\n",
      type, id, region, PrPt(C->gc.s) );
      PL( tmpbuf );
   }

   if( mask & gc_width )
   {
      Clr( C->gc, gc_width );
      sprintf( tmpbuf, "p %c \"%s\" %d gc width %g\n",
      type, id, region, C->gc.width );
      PL( tmpbuf );
   }

   if( mask & gc_type )
   {
      Clr( C->gc, gc_type );
      sprintf( tmpbuf, "p %c \"%s\" %d gc type %d\n",
      type, id, region, C->gc.type );
      PL( tmpbuf );
   }

   if( mask & gc_show )
   {
      Clr( C->gc, gc_show );
      sprintf( tmpbuf, "p %c \"%s\" %d gc show %d\n",
      type, id, region, C->gc.show );
      PL( tmpbuf );
   }

   return ERR_OK;
}

//------------------------------------------------------------------------------

int TclHInfo::DumpMP( HGContainerArc* ca, const char* id, int n )
{
   MN_Point mid;

   if( n <= 0 ) return ERR_OK;     // don't output arcs without middle points

   sprintf( tmpbuf, "p a \"%s\" 1 mp", id );
   PL( tmpbuf );
   for( int i = 0; i < n; i++ )
   {
      ca->AbsMP( i, mid );
      sprintf( tmpbuf, " %g %g", mid.x, mid.y );
      PL( tmpbuf );
   }
   PL( "\n" );

   return ERR_OK;
}

//------------------------------------------------------------------------------
