inline float ExprValue( Tcl_Interp* interp, char* expr )
{
   double d;
   Tcl_ExprDouble( interp, expr, &d );
   return (float)d;
}

typedef enum
{
   Direction_Fwd  = 0,
   Direction_Rev  = 1,
   Direction_None = 2
}
Direction;

//------------------------------------------------------------------------------

/*
class Entry {
public:
   HGObject*	o;
   int		t;
   FocalPt*	f;
   MN_Series	s;
};

class Stack {
public:
   Entry 	e[32];	// allow up to 32 entries (nesting levels)
   int last;		// corresponds to current nesting level
   Stack(void) { last = 0; }
   void Push(int ty) { e[last].t = ty; e[last].s.Clear(); last++; };
   void Pop(void)  { if( last > 0 ) last--; };
   void Set(HGObject* o) { if( last > 0 ) e[last-1].o = o; };
   void Set(FocalPt* f)  { if( last > 0 ) e[last-1].f = f; };
   void Set(float v)     { if( last > 0 ) e[last-1].s.Add( v ); };
};

extern Stack stack;
*/

//------------------------------------------------------------------------------

class TclTkHInfo : public HygraphInfo {
private:
   float*	matrix[16];
   MN_Series*	mseries[16];
   int		msize[16];

   int		UpdateSource( int indx, Vertex* u, int incr );
   int		UpdateSink( int indx, Vertex* u, int incr );

public:
   Tcl_Interp*	interp;
   VertexSet	fpts, sources, sinks;
   Tcl_HashTable *htable[3];	// used to look-up objs from widgets

		TclTkHInfo( const char* id, Tcl_Interp* interp );
   virtual	~TclTkHInfo( void ) {};

   int		ComputeStats( int type, int indx );

   int		CreateMatrix( int indx, int size );
   int		EnlargeMatrix( int indx, int size );
   float	Dist( int indx, bool normal );

   int		InitMinWtPaths( int indx, Direction d, int type,
		char* vvar, char* avar, char* vexpr, char* aexpr );
   float	MinWtPaths( int indx, Vertex* u, Vertex* v );
   int	 	MinWtPaths( int indx );
   int		OnPath( HGObject* o, int indx, int type );
   int		UpdateEndPt( Vertex* u, bool source, int incr );

   int		RegExpr( int type, int field, char* expr, int indx );
};

//------------------------------------------------------------------------------

class DQInfo {
public:
   unsigned long int	mask;		// bitmask wrt dyn query components
   bool			masked;		// is tuple masked?
   bool			painted;	// is tuple painted?
   float		val[32];	// attribute values
   char			color[16];	// actual color (when painting)
   int			i;		// index in vertex array (for APSP)
   char			type;		// type info: 0-v 1-a 2-b
   float		mag;		// focal point magnification

  // could restrict to 16 paths -> 2 paths/byte -> 8 bytes instead of 32!
   unsigned char	npaths[32];	// obj is on this # of shortest paths
					// in path object i

			DQInfo( int ty )
			{
			   type = ty;
			   mask = 0;
			   masked = false;
			   painted = false;
			   for( int i = 0; i < 32; i++ )
			   {
			      val[i] = 0;
			      npaths[i] = 0;
			   }
			};
};

#define DQ(FIELD,O)	(((DQInfo*)(O->vp))->FIELD)

//------------------------------------------------------------------------------

inline int Type( char t )
{
   switch( t )
   {
      case 'v':
      case 'f':
      case 's':
      case 'e':
                return 0;
      case 'a': return 1;
      case 'b': return 2;
      case 'h': return 3;
      default:  return -1;
   }
}

inline char Code( int t )
{
   switch( t )
   {
      case 0:  return 'v';
      case 1:  return 'a';
      case 2:  return 'b';
      case 3:  return 'h';
      default: return '?';
   }
}

//------------------------------------------------------------------------------
