//
// FILE:	tkHInfo.C
// AUTHORS:	Emanuel G. Noik
//
//		Tcl/Tk implementation of standard HGInfo methods.

#include <assert.h>
#include <tk.h>

#include <util/MN_Dictionary.h>
#include <util/MN_Set.h>
#include <util/MN_Point.h>
#include <util/unix.h>
#include <hygraph/Hygraph.h>
#include <server/io.h>
#include <server/server.h>
#include <server/hgInfo.h>

#include "tclTkHInfo.h"
#include "tkHInfo.h"

//------------------------------------------------------------------------------

// Private variables.
//
static char buf[512];

//------------------------------------------------------------------------------

static char* TkColor( MN_Point p )
{
   sprintf( buf, "#%02x%02x%02x",
      (int)(255.0 * p.x), (int)(255.0 * p.y), (int)(255.0 * p.z) );
   return buf;
}

//------------------------------------------------------------------------------

static char* TkCoords( HGContainer* C, MN_Point* pos, MN_Point* size,
int t, int region )
{
   switch( region )
   {
      case GEOM_ICON:
         //sprintf( buf, "%g %g", pos->x + size->x, pos->y + size->y );
         sprintf( buf, "%g %g %g %g", pos->x, pos->y, pos->x + size->x,
                  pos->y + size->y ); // anchor = nw
         break;
      case GEOM_LABEL:	// anchor = sw
         sprintf( buf, "%g %g", pos->x, pos->y + size->y );
         break;
      default:
         //sprintf( buf, "%g %g", pos->x, pos->y ); // anchor = nw
         sprintf( buf, "%g %g %g %g", pos->x, pos->y, pos->x + size->x,
                  pos->y + size->y ); // anchor = nw
         break;
   }
   return buf;
}

//------------------------------------------------------------------------------

static char* TkCoords( HGContainerArc* ca, int n )
{
   int f[4], t[4], ii[2], ij[2];
   //int tmp[4];
   MN_Point pt, size;
   HGContainer* fc = ca->From();
   HGContainer* tc = ca->To();

   //( fc->FindSubContainer( REG_I ) )->GetAbsPos( pt );
   //( fc->FindSubContainer( REG_I ) )->GetSize( size );
   fc->GetAbsPos( pt );
   fc->GetSize( size );
   f[0] = (int)pt.x;
   f[1] = (int)pt.y;
   f[2] = (int)( pt.x + size.x );
   f[3] = (int)( pt.y + size.y );
   //( tc->FindSubContainer( REG_I ) )->GetAbsPos( pt );
   //( tc->FindSubContainer( REG_I ) )->GetSize( size );
   tc->GetAbsPos( pt );
   tc->GetSize( size );
   t[0] = (int)pt.x;
   t[1] = (int)pt.y;
   t[2] = (int)( pt.x + size.x );
   t[3] = (int)( pt.y + size.y );

   ComputeAttachmentPoints( f, t, ii, ij );

   if( n <= 0 )
   {
      //ComputeAttachmentPoints( f, t, ii, ij );
      sprintf( buf, "%d %d %d %d", ii[0], ii[1], ij[0], ij[1] );
   }
   else
   {
      // set pt to the first middle point
      //
/*
      ca->AbsMP( 0, pt );
      tmp[0] = (int)pt.x - 1;
      tmp[1] = (int)pt.y - 1;
      tmp[2] = (int)pt.x + 1;
      tmp[3] = (int)pt.y + 1;
      ComputeAttachmentPoints( f, tmp, ii, ij );
*/

      sprintf( buf, "%d %d", ii[0], ii[1] );

      for( int i = 0; i < n; i++ )
      {
         ca->AbsMP( i, pt );
         sprintf( buf + strlen( buf ), " %d %d", (int)pt.x, (int)pt.y );
      }

      // pt is now set to the last middle point
      //
/*
      tmp[0] = (int)pt.x - 1;
      tmp[1] = (int)pt.y - 1;
      tmp[2] = (int)pt.x + 1;
      tmp[3] = (int)pt.y + 1;
      ComputeAttachmentPoints( tmp, t, ii, ij );
*/
      sprintf( buf + strlen( buf ), " %d %d", ij[0], ij[1] );
   }

   return buf;
}

//------------------------------------------------------------------------------

TkHInfo* TkAddH( const char* id, Tcl_Interp* interp )
{
   MN_String* s = new MN_String( id );
   TkHInfo* i = (TkHInfo*)hiDict->LookUp( s );

   if( i == NIL )
   {
      i = new TkHInfo( id, interp );
      i->hygraph->Validate( false );
      hiDict->Add( s, i );
   }
   else
      delete s;

   return i;
}

//------------------------------------------------------------------------------

int TkHInfo::DumpBB( HGContainer* C, const char* id, int t, int region )
{
   MN_Point pos, size;

   C->GetAbsPos( pos );
   C->GetSize( size );

   if( size.Magnitude() > 0.0 && C->widget != 0 )
   {
      sprintf( tmpbuf, ".canvas coords %d %s", C->widget,
         TkCoords( C, &pos, &size, t, region ) );
      Tcl_GlobalEval( interp, tmpbuf );
   }

   if( t == 3 )
   {
      sprintf( tmpbuf, "p h 0 1 bb %s",
         TkCoords( C, &pos, &size, t, region ) );
      Tcl_GlobalEval( interp, tmpbuf );
   }

   return ERR_OK; //DumpGC( C, id, t, region, gc_mask );
}

//------------------------------------------------------------------------------

int TkHInfo::DumpGC( HGContainer* C, const char* id, int t, int region, int m )
{
   int mask = C->gc.dirty & m;
   char type = Type( t );

   if( mask & gc_fill )
   {
      Clr( C->gc, gc_fill );
      sprintf( tmpbuf, "p %c \"%s\" %d gc fill %d\n",
      type, id, region, C->gc.fill );
      Tcl_GlobalEval( interp, tmpbuf );
   }

   if( mask & gc_fg_color )
   {
      Clr( C->gc, gc_fg_color );
      sprintf( tmpbuf, ".canvas configure %d -fill %s",
      C->widget, TkColor(C->gc.fg_color) );
      Tcl_GlobalEval( interp, tmpbuf );
   }

   if( mask & gc_bg_color )
   {
      Clr( C->gc, gc_bg_color );
      sprintf( tmpbuf, ".canvas configure %d -fill %s",
      C->widget, TkColor(C->gc.bg_color) );
      Tcl_GlobalEval( interp, tmpbuf );
   }

   if( mask & gc_width )
   {
      Clr( C->gc, gc_width );
      sprintf( tmpbuf, ".canvas configure %d -width %g",
      C->widget, C->gc.width );
      Tcl_GlobalEval( interp, tmpbuf );
   }

/*
   if( mask & gc_type )
   {
      Clr( C->gc, gc_type );
      sprintf( tmpbuf, "p %c \"%s\" %d gc type %d\n",
      type, id, region, C->gc.type );
      Tcl_GlobalEval( interp, tmpbuf );
   }

   if( mask & gc_show )
   {
      Clr( C->gc, gc_show );
      sprintf( tmpbuf, "p %c \"%s\" %d gc show %d\n",
      type, id, region, C->gc.show );
      Tcl_GlobalEval( interp, tmpbuf );
   }
*/

   return ERR_OK;
}

//------------------------------------------------------------------------------

int TkHInfo::DumpMP( HGContainerArc* ca, const char* id, int n )
{
   sprintf( tmpbuf, ".canvas coords %d %s",
      ca->widget, TkCoords( ca, n ) );
   Tcl_GlobalEval( interp, tmpbuf );
   return ERR_OK;
}

//------------------------------------------------------------------------------

void TkHInfo::Scale( MN_Point* s )
{
   if( s->x <= 0 ) s->x = 0.2;
   if( s->y <= 0 ) s->y = 0.2;
   s->z = 1;

   Vertex* v;
   for( int i = 0; (v = hygraph->V( i )) != 0; i++ )
   {
      MN_Point
         d( v->c->pos.x + v->c->offset.x, v->c->pos.y + v->c->offset.y, 0 );
      d = d * (*s) - d;
      Move( v, &d );
   }
}

//------------------------------------------------------------------------------

/*
void TkHInfo::Scale( MN_Point* s )
{
   if( s->x <= 0 ) s->x = 0.2;
   if( s->y <= 0 ) s->y = 0.2;
   s->z = 1;

   Scale( hygraph->c, s );
   Redisplay( hygraph->c );
}
*/

void TkHInfo::Scale( HGContainer* C, MN_Point* s )
{
   // Scale about (0,0).
   //
   if( C->gc.show )
   {
      C->offset *= *s;
   }
   else
   {
      MN_Point d( 10000, 10000, 0 );
      C->offset = (C->offset + d) * (*s) - d;
   }
   C->pos *= *s;
   C->size *= *s;

   for( HGContainer* c = C->Child(); c ; c = c->Sibling() )
      Scale( c, s );
}

//------------------------------------------------------------------------------

void TkHInfo::Redisplay( HGContainer* C )
{
   if( ! C->gc.show ) return;

   if( C->widget != 0 )
   {
      if( C->key == REG_L )
         sprintf( tmpbuf, ".canvas coords %d %g %g", C->widget,
            C->pos.x + C->offset.x, C->pos.y + C->offset.y + C->size.y );
      else
         sprintf( tmpbuf, ".canvas coords %d %g %g %g %g", C->widget,
            C->pos.x + C->offset.x,
            C->pos.y + C->offset.y,
            C->pos.x + C->offset.x + C->size.x,
            C->pos.y + C->offset.y + C->size.y );
      Tcl_GlobalEval( interp, tmpbuf );
   }

   HGContainerArc* ca;
   int i;
   for( i = 0; (ca = C->in_arcs[i]) != NIL; i++ )
      if( ca->From()->gc.show )
         DumpMP( ca, "", ca->MP() );
   for( i = 0; (ca = C->out_arcs[i]) != NIL; i++ )
      if( ca->To()->gc.show )
         DumpMP( ca, "", ca->MP() );

   for( HGContainer* c = C->Child(); c ; c = c->Sibling() )
      Redisplay( c );
}

//------------------------------------------------------------------------------

void TkHInfo::Translate( HGContainer* C, MN_Point* d )
{
   if( C->widget != 0 )
   {
//sprintf( tmpbuf, ".canvas coords %d", C->widget);
//Tcl_GlobalEval( interp, tmpbuf );
//printf("translate: old coords %s dx %g dy %g\n",interp->result,d->x,d->y);
      sprintf( tmpbuf, ".canvas move %d %g %g", C->widget, d->x, d->y );
      Tcl_GlobalEval( interp, tmpbuf );
//sprintf( tmpbuf, ".canvas coords %d", C->widget);
//Tcl_GlobalEval( interp, tmpbuf );
//printf("translate: new coords %s\n",interp->result);
   }
   for( HGContainer* c = C->Child(); c ; c = c->Sibling() )
      Translate( c, d );
}

//------------------------------------------------------------------------------

void TkHInfo::Move( Vertex* v, MN_Point* d )
{
   v->c->pos += *d;
   v->c->ComputeOffsets();

   //Translate( v->c, d );
   DumpGeom( v->c, "", geom[0], 0 );

   Move( v );
}

//------------------------------------------------------------------------------

void TkHInfo::Move( Vertex* v )
{
   Blob* b;
   AArc* a;
   int i;

   for( i = 0; (a = hygraph->In(v,i)) != NIL; i++ )
      if( ! DQ(masked,a) )
         DumpMP( a->ca, "", a->ca->MP() );
   for( i = 0; (a = hygraph->Out(v,i)) != NIL; i++ )
      if( ! DQ(masked,a) )
         DumpMP( a->ca, "", a->ca->MP() );
   for( i = 0; (b = hygraph->B(v,i)) != NIL; i++ )
      Move( b );
}

//------------------------------------------------------------------------------

void TkHInfo::Move( Blob* b )
{
   Vertex* u;
   for( int j = 0; (u = hygraph->To(b,j)) != NIL; j++ )
      Move( u );
}

//------------------------------------------------------------------------------
